"""Variation module."""

from .variation import (
    avdev,
    b_index,
    gibbs_m1,
    gibbs_m2,
    mcintosh_d,
    mod_vr,
    negative_extropy,
    range_vr,
    renyi_entropy,
)

__all__ = [
    "avdev",
    "b_index",
    "gibbs_m1",
    "gibbs_m2",
    "mcintosh_d",
    "mod_vr",
    "negative_extropy",
    "range_vr",
    "renyi_entropy",
]
