"""Dispersion module."""

from .dispersion import (
    coefficient_of_lvariation,
    coefficient_of_range,
    coefficient_of_variation,
    cole_index_of_dispersion,
    fisher_index_of_dispersion,
    gini_mean_difference,
    inter_expectile_range,
    morisita_index_of_dispersion,
    quartile_coefficient_of_dispersion,
    robust_coefficient_of_variation,
    shamos_estimator,
    standard_quantile_absolute_deviation,
    studentized_range,
)

__all__ = [
    "coefficient_of_lvariation",
    "coefficient_of_range",
    "coefficient_of_variation",
    "cole_index_of_dispersion",
    "fisher_index_of_dispersion",
    "gini_mean_difference",
    "inter_expectile_range",
    "morisita_index_of_dispersion",
    "quartile_coefficient_of_dispersion",
    "robust_coefficient_of_variation",
    "shamos_estimator",
    "standard_quantile_absolute_deviation",
    "studentized_range",
]
